{
  "nbformat": 4,
  "nbformat_minor": 0,
  "metadata": {
    "colab": {
      "provenance": []
    },
    "kernelspec": {
      "name": "python3",
      "display_name": "Python 3"
    },
    "language_info": {
      "name": "python"
    }
  },
  "cells": [
    {
      "cell_type": "markdown",
      "source": [
        "# オブジェクト指向回 演習問題"
      ],
      "metadata": {
        "id": "XQSw1mRcPbSf"
      }
    },
    {
      "cell_type": "markdown",
      "source": [
        "## 必要なmoduleのimport"
      ],
      "metadata": {
        "id": "M8RBjPeZRhxb"
      }
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "id": "rVPZLTePPY46"
      },
      "outputs": [],
      "source": [
        "from abc import ABC, abstractmethod\n",
        "from typing import Literal, Optional\n",
        "from typing_extensions import override"
      ]
    },
    {
      "cell_type": "markdown",
      "source": [
        "## Entityの定義"
      ],
      "metadata": {
        "id": "iBQxEcNHRmGl"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "class Entity:\n",
        "    def __init__(self, is_enemy: bool, name: str, hp: int, attack: int):\n",
        "        self.__is_enemy = is_enemy\n",
        "        self.__name = name\n",
        "        self.__hp = hp\n",
        "        self.__attack = attack\n",
        "\n",
        "    @property\n",
        "    def is_enemy(self):\n",
        "        return self.__is_enemy\n",
        "\n",
        "    @property\n",
        "    def name(self):\n",
        "        return self.__name\n",
        "\n",
        "    @property\n",
        "    def hp(self):\n",
        "        return self.__hp\n",
        "\n",
        "    @hp.setter\n",
        "    def hp(self, value: int):\n",
        "        # 負の数にならないようにする\n",
        "        self.__hp = max(0, value)\n",
        "\n",
        "    @property\n",
        "    def attack(self):\n",
        "        return self.__attack\n",
        "\n",
        "    def commands(self):\n",
        "        return [AttackCommand()]\n",
        "\n",
        "\n",
        "class Battle:\n",
        "    enemies: list[Entity]\n",
        "    players: list[Entity]\n",
        "\n",
        "    def __init__(self, players: list[Entity], enemies: list[Entity]):\n",
        "        self.players = players\n",
        "        self.enemies = enemies\n",
        "\n",
        "    def print_status(self):\n",
        "        for player in self.players:\n",
        "            print(f\"自 {player.name}: HP={player.hp}\")\n",
        "        for enemy in self.enemies:\n",
        "            print(f\"敵 {enemy.name}: HP={enemy.hp}\")\n",
        "        print(\"----------\")\n",
        "\n",
        "    def is_settled(self) -> Optional[Literal[\"player\", \"enemies\"]]:\n",
        "        if all(enemy.hp <= 0 for enemy in self.enemies):\n",
        "            return \"player\"\n",
        "        if all(player.hp <= 0 for player in self.players):\n",
        "            return \"enemies\"\n",
        "        return None\n",
        "\n",
        "    def all_entities(self):\n",
        "        return [*self.players, *self.enemies]\n",
        "\n",
        "    def start(self):\n",
        "        turn = \"player\"\n",
        "        while not (judgement := self.is_settled()):\n",
        "            self.print_status()\n",
        "            if turn == \"player\":\n",
        "                self.player_turn()\n",
        "                turn = \"enemies\"\n",
        "            else:\n",
        "                self.enemies_turn()\n",
        "                turn = \"player\"\n",
        "\n",
        "        if judgement == \"player\":\n",
        "            print(\"プレイヤーの勝ち\")\n",
        "        else:\n",
        "            print(\"プレイヤーの負け\")\n",
        "\n",
        "    def player_turn(self):\n",
        "        # コマンドを表示\n",
        "        decision = []\n",
        "        for player in self.players:\n",
        "            if player.hp == 0:\n",
        "                continue\n",
        "\n",
        "            print(f\"{player.name}のターン\")\n",
        "            commands = player.commands()\n",
        "            for id, command in enumerate(commands):\n",
        "                print(f\"{id}: {command.name}\")\n",
        "\n",
        "            # コマンドを選択\n",
        "            command = commands[ask_int(\"どうする？> \", 0, len(commands) - 1)]\n",
        "            print(\"----------\")\n",
        "\n",
        "            # ターゲットを選択\n",
        "            all_entities = self.all_entities()\n",
        "            targets = command.possible_targets(all_entities, player)\n",
        "            for id, target in enumerate(targets):\n",
        "                print(f\"{id}: {target.name} HP={target.hp}\")\n",
        "            target = targets[ask_int(\"誰に？> \", 0, len(targets) - 1)]\n",
        "            print(\"----------\")\n",
        "            decision.append((player, command, target))\n",
        "\n",
        "        for player, command, target in decision:\n",
        "            command.execute(player, target)\n",
        "\n",
        "    def enemies_turn(self):\n",
        "        # とりあえず敵は攻撃するだけ\n",
        "        for enemy in self.enemies:\n",
        "            if enemy.hp == 0:\n",
        "                continue\n",
        "            # 生存している最初のプレイヤーに攻撃\n",
        "            player = next(player for player in self.players if player.hp > 0)\n",
        "            enemy.commands()[0].execute(enemy, player)\n",
        "\n",
        "def ask_int(prompt: str, min_value: int, max_value: int) -> int:\n",
        "    while True:\n",
        "        try:\n",
        "            value = int(input(prompt))\n",
        "            if min_value <= value <= max_value:\n",
        "                return value\n",
        "        except (KeyboardInterrupt, EOFError):\n",
        "            raise\n",
        "        except:\n",
        "            pass\n",
        "        print(\"範囲内の数字を入力してください\")\n",
        "\n",
        "\n",
        "# プレイヤー\n",
        "class Player(Entity):\n",
        "    def __init__(self, name: str, hp: int, attack: int):\n",
        "        super().__init__(False, name, hp, attack)\n",
        "\n",
        "    def commands(self):\n",
        "        return [AttackCommand(), HealCommand()]\n",
        "\n",
        "\n",
        "# 敵\n",
        "class EnemySlime(Entity):\n",
        "    def __init__(self):\n",
        "        super().__init__(True, \"スライム\", 10, 1)\n",
        "\n",
        "\n",
        "class EnemyGoblin(Entity):\n",
        "    def __init__(self):\n",
        "        super().__init__(True, \"ゴブリン\", 20, 5)\n"
      ],
      "metadata": {
        "id": "JObAEuApPc0t"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "source": [
        "## コマンドの定義"
      ],
      "metadata": {
        "id": "l4nYxBYLRooA"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "class Command(ABC):\n",
        "    def __init__(self, name: str):\n",
        "        self.__name = name\n",
        "\n",
        "    @property\n",
        "    def name(self):\n",
        "        return self.__name\n",
        "\n",
        "    @abstractmethod\n",
        "    def possible_targets(self, entities: list[Entity], operator: Entity):\n",
        "        \"このコマンドの対象となりうるエンティティを列挙する\"\n",
        "        ...\n",
        "\n",
        "    @abstractmethod\n",
        "    def execute(self, operator: Entity, target: Entity):\n",
        "        \"operatorがこのコマンドをtargetに対して実行する\"\n",
        "        ...\n",
        "\n",
        "class AttackCommand(Command):\n",
        "    def __init__(self):\n",
        "        super().__init__(\"攻撃\")\n",
        "\n",
        "    @override\n",
        "    def possible_targets(self, entities: list[Entity], operator: Entity):\n",
        "        # 自分と違うサイドのエンティティを攻撃可能\n",
        "        return [e for e in entities if e.is_enemy != operator.is_enemy]\n",
        "\n",
        "    @override\n",
        "    def execute(self, operator: Entity, target: Entity):\n",
        "        print(f\"{operator.name}の攻撃！\")\n",
        "        print(f\"{target.name}に{operator.attack}のダメージ\")\n",
        "        target.hp -= operator.attack\n",
        "\n",
        "\n",
        "class HealCommand(Command):\n",
        "    def __init__(self):\n",
        "        super().__init__(\"回復\")\n",
        "\n",
        "    @override\n",
        "    def possible_targets(self, entities: list[Entity], operator: Entity):\n",
        "        # 自分自身のみ回復可能\n",
        "        return [operator]\n",
        "\n",
        "    @override\n",
        "    def execute(self, operator: Entity, target: Entity):\n",
        "        print(f\"{operator.name}の回復！\")\n",
        "        print(f\"{target.name}のHPが10回復した\")\n",
        "        target.hp += 10\n"
      ],
      "metadata": {
        "id": "-UwwyrZhPc29"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "source": [
        "## プレイ"
      ],
      "metadata": {
        "id": "oG-8QXxdRuxt"
      }
    },
    {
      "cell_type": "code",
      "source": [
        "players: list[Entity] = [\n",
        "    Player(\"ゆうしゃ\", 30, 10),\n",
        "]\n",
        "enemies: list[Entity] = [\n",
        "    EnemySlime(),\n",
        "    EnemyGoblin(),\n",
        "]\n",
        "battle = Battle(players, enemies)\n",
        "battle.start()"
      ],
      "metadata": {
        "id": "LkZudv-GPc5k"
      },
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "source": [],
      "metadata": {
        "id": "QxwL9iHWPc60"
      },
      "execution_count": null,
      "outputs": []
    }
  ]
}